# -*- coding: utf-8 -*-

import numpy as np
from models.GPModel import GPModel
from acquisition.ei import EI
from acquisition.ei_mc import EIMC
from acquisition.lcb import LCB
from acquisition.ucb import UCB
from optimizer.LBFGSB import LBFGSB
from optimizer.SLSQP import SLSQP
from optimizer.DE import DE
from variable.continuous_variable import ContinuousVariable
from variable.discrete_variable import DiscreteVariable
from variable.categorical_variable import CategoricalVariable
from points_generator.bruteforce_points_generator import BruteforcePointsGenerator
from points_generator.random_points_generator import RandomPointsGenerator
from tqdm import tqdm
from .evaluators.batch_sequential import BatchSequential
from .evaluators.batch_thompson import ThompsonBatch
from utils.plot import plot_acquisition, plot_approximation
from utils.math import normalize_X, denormalize_X, normalize_Y_f, one_hot_to_cat, is_number
import matplotlib.pyplot as plt
import copy
import csv
from datetime import datetime
import os



class BayesianOptimization(object):
    def __init__(self, domain_x, objectives, domain_y=None, g=None, constraints=None, acquisition_type='EI', 
                 acquisition_optimizer_type='lbfgsb', model_type='GP', kernel=None, 
                 X_init=None, Y_init=None, init_n_samples=5, samples_generator='random', 
                 normalize_X=False, normalize_Y=False, ARD=True, 
                 evaluator_type='sequential_batch', batch_size=1, save_experiment=False, load_experiment_foldername=None, **kwargs):
        
        if len(domain_y)!=len(objectives):
            raise Exception("Cardinality of domain_y doesn't match number of objectives")
            
        self.model_type=model_type
        self.kernel = kernel
        self.dictionary_X=copy.deepcopy(domain_x)
        self.domain_x=self._init_domain(domain_x)
        self.domain_y=domain_y
        
        # split variable in categorical and non-categorical and fix domain_x
        self.fvars_cat, self.fvars_not_cat = self._split_variables(self.domain_x)
        self.domain_x = self.fvars_not_cat + self.fvars_cat
        
        self.objectives=objectives
        self.constraints=constraints
        self.acquisition_type=acquisition_type
        self.acquisition_optimizer_type=acquisition_optimizer_type
        self.g = g
        self.X=X_init
        self.Y=Y_init
        self.init_n_samples=init_n_samples
        self.samples_generator_type=samples_generator
        self.normalize_X=normalize_X
        self.normalize_Y=normalize_Y
        self.ARD=ARD
        self.batch_size=batch_size
        self.kwargs=kwargs
        self.load_experiment_foldername=load_experiment_foldername
        
        # Set up initial X and Y. Generate samples if not provided
        self._init_design_chooser()      
        self._initialize_bounds()
        self.model = [self._model_chooser(copy.deepcopy(self.kernel)) for ys in range(len(self.domain_y))]
        self.acquisition = self._acquisition_chooser()
        self.acquisition_optimizer = self._acquisition_optimizer_chooser()
        self.samples_generator = self._sample_generator_chooser()
        
        
        if load_experiment_foldername!=None:
            self._load_experiment()
        
        self.one_hot_permutations=None
        if self.samples_generator_type == 'bruteforce':
            self.one_hot_permutations, self.cat_permutations = self.samples_generator.generate_points_cat_permutations()
        self.evaluator_type=evaluator_type
        self.evaluator=self._evaluator_chooser()
        self.save_experiment=save_experiment
        
        
    def _init_domain(self, domain_x):
        variables_list=[]
        for function_var in domain_x:
            if function_var['type']=='continuous':
                variables_list.append(ContinuousVariable(function_var['name'], np.array(function_var['domain'])))
            elif function_var['type']=='discrete':
                variables_list.append(DiscreteVariable(function_var['name'], np.array(function_var['domain'])))
            elif function_var['type']=='categorical':
                variables_list.append(CategoricalVariable(function_var['name'], np.array(function_var['domain'])))
            else:
                raise Exception('Variable type not recognized:'+function_var['type'])
        return variables_list
    
    def _split_variables(self, fvars):
        # Separate categorical from others
        fvars_cat = []
        fvars_not_cat = []
        for fvar in fvars:
            if fvar.get_type() == 'categorical':
                fvars_cat.append(fvar)
            else:
                fvars_not_cat.append(fvar)
        return fvars_cat, fvars_not_cat    
    
    def _generate_samples(self):
        X_init = np.empty((self.init_n_samples, 0))
        # each iteration adds a dimension(feature) to all initial points
        for function_var in self.domain_x:
            if function_var.get_type() == 'continuous':
                x = np.random.uniform(function_var.get_domain()[0], function_var.get_domain()[1], (self.init_n_samples,1))
            elif function_var.get_type() == 'discrete':
                x = np.random.choice(function_var.get_domain(), (self.init_n_samples,1))
            elif function_var.get_type() == 'categorical':
                x = function_var.get_one_hot(np.random.choice(function_var.get_domain(), (self.init_n_samples,1)))
            else:
                raise ValueError('Variable type not supported')
            X_init = np.column_stack((X_init, x))  
        return X_init
        
    def _init_design_chooser(self):
        if self.load_experiment_foldername!=None:
            self.Y=[np.empty((0,1), dtype=np.float64) for i in range(len(self.domain_y))]
            self.X=np.empty((0,len(self.dictionary_X)))
            return
        if self.X is None:
            self.X=self._generate_samples()
                
        if self.Y is None:
            self.Y = []
            for idx, objectives in enumerate(self.objectives):
                # is it better to have an objective class and do objective.evaluate(X)?
                if self.fvars_cat!=[]:
                    self.Y.append(objectives(one_hot_to_cat(self.domain_x, self.X)).reshape(-1,1))
                else:
                    self.Y.append(objectives(self.X).reshape(-1,1))
    
    
    def _evaluator_chooser(self):
        if self.evaluator_type=='sequential_batch':
            return BatchSequential(self.acquisition, self.batch_size, self.domain_x,
                                   self.normalize_X, self.samples_generator_type, 
                                   self.one_hot_permutations, self.acquisition_optimizer)
        elif self.evaluator_type=='thompson':
            return ThompsonBatch(self.acquisition, self.batch_size, self.domain_x,
                                 self.normalize_X, self.samples_generator_type, 
                                 self.one_hot_permutations, self.acquisition_optimizer)
        else:
            raise Exception("Not yet implemented")
    
    def _initialize_bounds(self):
        bounds = np.empty((0,2))
        if self.samples_generator_type == 'bruteforce':
            for function_var in self.fvars_not_cat:
                bounds = np.vstack((bounds, function_var.get_bounds()))
        if self.samples_generator_type == 'random':
            for function_var in self.domain_x:
                bounds = np.vstack((bounds, function_var.get_bounds()))
        self.bounds=bounds
        
    def _model_chooser(self, kernel):
        # in GPyOpt this part is in util/arguments_manager
        if self.model_type=='GP':
            #kernel=self.kwargs.get('kernel', None)
            ARD=self.ARD
            optimize_restarts=self.kwargs.get('optimize_restarts', 5)
            model_optimizer_type = self.kwargs.get('model_optimizer_type','lbfgs')
            max_iters=self.kwargs.get('max_iters', 1000)
            return GPModel(kernel, model_optimizer_type, optimize_restarts, max_iters, ARD, normalize_Y=self.normalize_Y)
        
    def _update_model(self, model, X, Y, idx_model):
        X_in_model=[]
        Y_in_model=[]
        for i in range(X.shape[0]):
            if is_number(Y[i]):
                X_in_model.append(X[i])
                Y_in_model.append(Y[i])
        X_in_model=np.asarray(X_in_model, dtype=np.float64).reshape(-1, X.shape[1])
        Y_in_model=np.asarray(Y_in_model, dtype=np.float64).reshape(-1, 1)
        if self.normalize_X:
            X_in_model = normalize_X(self.domain_x, X_in_model)

            
        #self.model.updateModel(X_in_model, Y_in_model)
        model.updateModel(X_in_model, Y_in_model)

    
    def _acquisition_chooser(self):
        if self.acquisition_type!='EIMC' and len(self.objectives)!=1:
            raise Exception("Multi output only supported with EIMC")
        if self.acquisition_type=='EI':
            #return EI(self.model)
            return EI()
        elif self.acquisition_type=='EIMC':
            # identity function if none
            if self.g == None and len(self.objectives) == 1:
                def g(x, y): return y
                self.g = g
            elif self.g == None:
                raise Exception("g cannot be empty for multi output functions")
            return EIMC(self.g)
        elif self.acquisition_type=='LCB':
            return LCB()
        elif self.acquisition_type=='UCB':
            return UCB()
        else:
            raise Exception("acq f not coded")                
            
    def _acquisition_optimizer_chooser(self):
        if self.acquisition_optimizer_type=='lbfgsb':
            opt=LBFGSB()
        elif self.acquisition_optimizer_type=='de':
            opt=DE()
        elif self.acquisition_optimizer_type=='slsqp':
            opt=SLSQP()
        else:
            raise Exception("optimizer unknown")
        
        # If normalize, then normalize even the boundaries
        if self.normalize_X:
            norm_bounds = np.zeros_like(self.bounds)
            norm_bounds[:,1] = np.ones((self.bounds.shape[0]))
            opt.set_bounds(norm_bounds)
        else:
            opt.set_bounds(self.bounds)
        return opt
    
    def _sample_generator_chooser(self):
        if self.samples_generator_type == 'random':
            return RandomPointsGenerator(self.domain_x)
        if self.samples_generator_type == 'bruteforce':
            return BruteforcePointsGenerator(self.fvars_not_cat, self.fvars_cat)
            
    def _get_rounded_opt(self, X):
        X_round = np.zeros_like(X)
        idx = 0
        for fvar in self.domain_x:
            dim = fvar.get_dim()
            X_round[:,idx:idx+dim] = fvar.get_rounded(X[:,idx:idx+dim])
            idx = idx + dim
        return X_round
    
    
    
    def run_optimization(self, max_iter=10, max_time = np.inf, plot=False, eps = 1e-8):
        if self.load_experiment_foldername==None:
            self.folder_name="Experiment"+datetime.now().strftime("%Y%m%d-%H%M%S")
        else:
            self.folder_name=self.load_experiment_foldername
        
        if plot==True:
            plt.figure(figsize=(10, np.maximum(1,int(max_iter/5)) * 3))
            plt.subplots_adjust(hspace=0.4)
            plt.ylim([-5,5])
            plot_counter = 0
            
        '''
        self.model is always a list of models. In this way the constructor of
        BayesianOptimization is compatible both with EIMC(where g() is a combination
        of the Ys(called objectivess) and the others acquisitions(where g()
        simply returns Y)
        The following 'if condition' has been placed so that the non-EIMC
        acquisitions, which use a single GP, don't handle a list but a single
        GP model.
        '''
        if self.acquisition_type!='EIMC':
            self.acquisition.set_model(self.model[0])
        else:
            self.acquisition.set_model(self.model)
            
        #self.warning = False
        
        for iteration in tqdm(range(max_iter)):
            
            self.acquisition.set_constraints(self.constraints)
            if self.acquisition_type=='EIMC':
                self.acquisition.set_all_X(self.X)
            
            for idx, model in enumerate(self.model):
                self._update_model(model, self.X, self.Y[idx], idx)
            
            X_next = self.evaluator.compute_batch()
            
            if self.normalize_X:
                X_next = denormalize_X(self.domain_x, X_next)
              
            Y_next=[]
            
            # Calculate y for every objectives.
            for idx, _ in enumerate(self.objectives):
                if self.fvars_cat != []:
                    Y_next.append(self.objectives[idx](one_hot_to_cat(self.domain_x, X_next)))
                else:
                    Y_next.append(self.objectives[idx](X_next))

            #if plot==True:
            #    plt.subplot(max_iter, 2, 2*iteration+1)
            #    Y_to_plot=self.Y[0] if self.normalize_Y==False else normalize_Y_f(self.Y[0])
            #    plot_approximation(model, self.X, Y_to_plot, X_next, self.domain_x, self.normalize_X, 
            #                       self.constraints, show_legend=iteration==0)
            #    plt.title(f'Iteration {iteration+1}')
            #    
            #    plt.subplot(max_iter, 2, 2*iteration+2)
            #    plot_acquisition(model, self.domain_x, self.bounds, self.acquisition, X_next, self.normalize_X, 
            #                     self.constraints)
            
            if plot==True and (iteration % 5) == 0:
                plot_counter += 1
                plt.subplot(np.maximum(1,int(max_iter/5)), 2, plot_counter)
                Y_to_plot = self.Y[0]
                plot_approximation(model, self.X, Y_to_plot, X_next, self.domain_x, self.normalize_X, 
                                   self.constraints, show_legend=iteration==0) 
                plt.title('Surrogate function \n (iteration: {0})'.format(iteration))
                plot_counter += 1
                plt.subplot(np.maximum(1,int(max_iter/5)), 2, plot_counter)
                plot_acquisition(model, self.domain_x, self.bounds, self.acquisition, X_next, self.normalize_X, 
                                 self.constraints)
                plt.title('Acquisition function \n (iteration: {0})'.format(iteration))
                plt.tight_layout()
            
            if self.save_experiment:
                self._save_experiment(X_next, Y_next)
            
            # Add sample to previous samples
            self.X = np.vstack((self.X, X_next))
            for idx, _ in enumerate(self.Y):
                self.Y[idx] = np.vstack((self.Y[idx], Y_next[idx]))
            
            #if (-self.acquisition.get_acquisition(X_next)) < eps:
            #    if self.warning:
            #        break
            #    else:
            #        self.warning = True
            #else:
            #    self.warning = False
            
            #if self._distance_last_evaluations() < eps:
            #    print("Stopping criteria reached: distance between last evaluations <{}".format(eps))
            #    break
            
        # Transform Y from list of np.array to only np.array
        Y_as_array=np.empty((self.X.shape[0], 0))
        for column in self.Y:
            Y_as_array=np.column_stack((Y_as_array, column))
        
        if plot == True:
            plt.figure(figsize=(10,5))
            plt.subplot(1, 2, 1)
            plot_approximation(model, self.X, Y_as_array, X_next, self.domain_x, self.normalize_X, 
                               self.constraints, show_legend=True)
            plt.subplot(1, 2, 2)
            plot_acquisition(model, self.domain_x, self.bounds, self.acquisition, X_next, self.normalize_X, 
                             self.constraints)
            plt.tight_layout()
        
        for idx, model in enumerate(self.model):
            self._update_model(model, self.X, self.Y[idx], idx)
        
        self.Y = Y_as_array
        
        
    def compute_results(self):
        if self.acquisition_type=='EIMC':
            best_g=np.inf
            for i in range(self.X.shape[0]):
                if not np.isnan(np.max(self.Y[i])):
                    g=self.g(np.atleast_2d(self.X[i]), np.atleast_2d(self.Y[i]))
                    if g<best_g:
                        best_index=i
                        best_g=g
            print("best x:{}, best y:{}, best g:{}".format(self.X[best_index], self.Y[best_index], best_g))
        else:
            best_y=np.inf
            for i in range(self.X.shape[0]):
                if not np.isnan(np.max(self.Y[i])):
                    y=self.objectives[0](np.atleast_2d(self.X[i]))
                    if y<best_y:
                        best_index=i
                        best_y=y
            print("best x:{}, best y:{}".format(self.X[best_index], self.Y[best_index]))
            
    def _load_experiment(self):
        with open(self.load_experiment_foldername+"/Data.csv") as filedata:
            csv_reader = csv.DictReader(filedata, delimiter=',')
            self.imported_X=[[] for i in range(len(self.domain_y))]
            self.imported_Y=[[] for i in range(len(self.domain_y))]
            all_X=[]
            for row in csv_reader:
                X_to_add=[]
                for fvar in self.dictionary_X:
                    X_to_add.append(row[fvar['name']])
                for idx, fvar in enumerate(self.domain_y):
                    if is_number(row[fvar['name']]):
                        self.imported_Y[idx].append(row[fvar['name']])
                    else:
                        self.imported_Y[idx].append(np.nan)
                all_X.append(X_to_add)
            # Convert to np array
            for i in range(len(self.domain_y)):
                self.Y[i]=np.asarray(self.imported_Y[i], dtype=np.float64).reshape(-1, 1)
            self.X=np.vstack((self.X, np.asarray(all_X, dtype=np.float64).reshape(-1, len(self.dictionary_X))))
                
    def _save_experiment(self, X_next=None, Y_next=None):
        folder_exists=os.path.exists(self.folder_name)
        if not folder_exists:
            os.makedirs(self.folder_name)
        with open(self.folder_name+'/Data.csv', mode='a', newline='') as file:
            csv_writer=csv.writer(file)
            
            if not folder_exists:
                header=[fvar['name'] for fvar in self.dictionary_X]
                header.extend([fvar['name'] for fvar in self.domain_y])
                csv_writer.writerow(header)
                matrix_to_write=np.empty((self.X.shape[0], self.X.shape[1]+len(self.objectives)))
                matrix_to_write[:,:self.X.shape[1]]=self.X[:,:]
                for i in range(matrix_to_write.shape[0]):
                    for j in range(self.X.shape[1], matrix_to_write.shape[1]):
                        matrix_to_write[i][j]=self.Y[j-self.X.shape[1]][i]
                csv_writer.writerows(matrix_to_write)

                #This is horrible, let's change Y in a np.array instead of list
                #csv_writer.writerow(X_next[0].tolist()+[y[0] for y in Y_next])
            matrix_to_write=np.empty((X_next.shape[0], X_next.shape[1]+len(self.objectives)))
            matrix_to_write[:,:X_next.shape[1]]=X_next[:,:]
            for i in range(matrix_to_write.shape[0]):
                for j in range(X_next.shape[1], matrix_to_write.shape[1]):
                    matrix_to_write[i][j]=Y_next[j-X_next.shape[1]][i]
            csv_writer.writerows(matrix_to_write)
            
    def _distance_last_evaluations(self):
        """
        Distance between the last two evaluations(X).
        """
        if self.X.shape[0] < 2:
            return np.inf
        return np.sqrt(np.sum((self.X[-1, :] - self.X[-2, :]) ** 2))
        