#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jan 20 21:26:26 2019

@author: Stefano Toniolo
"""

#%% Constraints

import core.BO as BO
import numpy as np
import GPy.kern as k

def six_humps(x):
    ''' f(x,y) = (4 - 2.1x^2 + x^4 / 3)x^2 + xy + (-4 + 4x^2)x^2 '''
    x2 = np.power(x[:,0],2)
    x4 = np.power(x[:,0],4)
    y2 = np.power(x[:,1],2)  
    fval = (4.0 - 2.1 * x2 + (x4 / 3.0)) * x2 + x[:,0]*x[:,1] + (-4.0 + 4.0 * y2) * y2
    return fval.reshape(-1,1)

def c1(x):
    return -x[:,1] -.5 + np.abs(x[:,0]) - np.sqrt(1-x[:,0]**2)

def c2(x):
    return x[:,1] +.5 - np.abs(x[:,0]) - np.sqrt(1-x[:,0]**2)

constraints = [{'type': 'ineq', 'fun': c1},
               {'type': 'ineq', 'fun': c2}]

fvars = [{'name':'var1', 'type':'continuous', 'domain': (-1,1)},
         {'name':'var2', 'type':'continuous', 'domain': (-2,2)}]

fvarsY = [{'name':'y1', 'type':'continuous', 'domain': (-100,100)}]

kernel = k.Matern52(input_dim=2)# + k.Bias(input_dim=1)
fs=[six_humps]

Xinit = np.array([[-0.2941242 ,  0.74985241],
                  [ 0.16504143,  0.42840476],
                  [-0.17331062,  0.65814185],
                  [ 0.97909193,  0.27580151],
                  [ 0.24891458,  0.71742811]])


Yinit = six_humps(Xinit)

# Create the BayesianOptimization object
my_bo=BO.BayesianOptimization(fvars, fs, fvarsY, normalize_Y=True, normalize_X=False, X_init=Xinit, #Y_init=Yinit,
                              acquisition_type='EI', acquisition_optimizer_type='slsqp',
                              init_n_samples=30, samples_generator='random', 
                              ARD=True, kernel=kernel, constraints=constraints)

my_bo.run_optimization(max_iter=25,plot=True)
print("x:",my_bo.X[np.argmin(my_bo.Y)]," y:",np.min(my_bo.Y))

#%% GPyOpt

import GPyOpt
import GPy
import numpy as np
import matplotlib.pyplot as plt

func = GPyOpt.objective_examples.experiments2d.sixhumpcamel()

space =[{'name': 'var_1', 'type': 'continuous', 'domain': (-1,1)},
        {'name': 'var_2', 'type': 'continuous', 'domain': (-1.5,1.5)}]

constraints = [{'name': 'constr_1', 'constraint': '-x[:,1] -.5 + abs(x[:,0]) - np.sqrt(1-x[:,0]**2)'},
              {'name': 'constr_2', 'constraint': 'x[:,1] +.5 - abs(x[:,0]) - np.sqrt(1-x[:,0]**2)'}]

feasible_region = GPyOpt.Design_space(space = space, constraints = constraints)

## Grid of points to make the plots
grid = 400
bounds = feasible_region.get_continuous_bounds()
X1 = np.linspace(bounds[0][0], bounds[0][1], grid)
X2 = np.linspace(bounds[1][0], bounds[1][1], grid)
x1, x2 = np.meshgrid(X1, X2)
X = np.hstack((x1.reshape(grid*grid,1),x2.reshape(grid*grid,1)))

## Check the points in the feasible region.
masked_ind = feasible_region.indicator_constraints(X).reshape(grid,grid)
masked_ind = np.ma.masked_where(masked_ind > 0.5, masked_ind)
masked_ind[1,1]=1

## Make the plots
plt.figure(figsize=(14,6))

# Feasible region
plt.subplot(121)
plt.contourf(X1, X2, masked_ind ,100, cmap= plt.cm.bone, alpha=1,origin ='lower')
plt.text(-0.25,0,'FEASIBLE',size=20)
plt.text(-0.3,1.1,'INFEASIBLE',size=20,color='white')

plt.subplot(122)
plt.plot()
plt.contourf(X1, X2, func.f(X).reshape(grid,grid),100, alpha=1,origin ='lower')
plt.plot(np.array(func.min)[:,0], np.array(func.min)[:,1], 'r.', markersize=20, label=u'Minimum')
plt.legend()
plt.title('Six-Hump Camel function',size=20)

plt.figure(figsize=(6.5,6))
OB = plt.contourf(X1, X2, func.f(X).reshape(grid,grid),100,alpha=1)
IN = plt.contourf(X1, X2, masked_ind ,100, cmap= plt.cm.bone, alpha=.5,origin ='lower')
plt.text(-0.25,0,'FEASIBLE',size=20,color='white')
plt.text(-0.3,1.1,'INFEASIBLE',size=20,color='white')
plt.plot(np.array(func.min)[:,0], np.array(func.min)[:,1], 'r.', markersize=20, label=u'Minimum')
plt.title('Six-Hump Camel with restrictions',size=20)
plt.legend()

# --- CHOOSE the intial design
from numpy.random import seed # fixed seed
seed(123456)

initial_design = GPyOpt.experiment_design.initial_design('random', feasible_region, 10)

plt.figure(figsize=(6.5,6))
OB = plt.contourf(X1, X2, func.f(X).reshape(grid,grid),100,alpha=1)
IN = plt.contourf(X1, X2, masked_ind ,100, cmap= plt.cm.bone, alpha=.5,origin ='lower')
plt.text(-0.25,0,'FEASIBLE',size=20,color='white')
plt.text(-0.3,1.1,'INFEASIBLE',size=20,color='white')
plt.plot(np.array(func.min)[:,0], np.array(func.min)[:,1], 'r.', markersize=20, label=u'Minimum')
plt.title('Six-Hump Camel with restrictions',size=20)
plt.plot(initial_design[:,0],initial_design[:,1],'yx',label = 'Design')
plt.legend()

# --- CHOOSE the objective
objective = GPyOpt.core.task.SingleObjective(func.f)

# --- CHOOSE the model type
model = GPyOpt.models.GPModel(exact_feval=True,optimize_restarts=10,verbose=False)

# --- CHOOSE the acquisition optimizer
aquisition_optimizer = GPyOpt.optimization.AcquisitionOptimizer(feasible_region)

# --- CHOOSE the type of acquisition
acquisition = GPyOpt.acquisitions.AcquisitionEI(model, feasible_region, optimizer=aquisition_optimizer)

# --- CHOOSE a collection method
evaluator = GPyOpt.core.evaluators.Sequential(acquisition)

# BO object
bo = GPyOpt.methods.ModularBayesianOptimization(model, feasible_region, objective, acquisition, evaluator, initial_design)

# --- Stop conditions
max_time  = None 
max_iter  = 5
tolerance = 1e-8     # distance between two consecutive observations  

# Run the optimization                                                  
bo.run_optimization(max_iter = max_iter, max_time = max_time, eps = tolerance, verbosity=False) 
bo.plot_acquisition()

# Run the optimization  
max_iter  = 25
bo.run_optimization(max_iter = max_iter, max_time = max_time, eps = tolerance, verbosity=False) 

bo.plot_acquisition()
bo.plot_convergence()

# Best found value
np.round(bo.x_opt,2)

# True min
np.round(func.min[0],2)