#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 30 11:59:30 2018

@author: Stefano Toniolo
"""

import sys
sys.path.append("..") # Adds higher directory to python modules path.

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from tqdm import tqdm

import GPy
from acquisition.ei import EI
from optimizer.optimizer import OptimizerLBRGS
from utils.plot import plot_approximation_2d, plot_acquisition
from utils.math import normalize
from core.variable import ContinuousVariable

import seaborn as sns
sns.set_style("white")

#%%

def f(x, y):
    ''' f(x,y) = (4 - 2.1x^2 + x^4 / 3)x^2 + xy + (-4 + 4x^2)x^2 '''
    x2 = np.power(x,2)
    x4 = np.power(x,4)
    y2 = np.power(y,2)  
    fval = (4.0 - 2.1 * x2 + (x4 / 3.0)) * x2 + x*y + (-4.0 + 4.0 * y2) * y2
    return -fval

#%%

var1 = ContinuousVariable('var1', [-2.0,2.0])
var2 = ContinuousVariable('var2', [-1.0,1.0])

#bounds = np.array([[-2.0, 2.0],     # x
#                   [-1.0, 1.0]])    # y

init_samples = 10

X1 = np.random.uniform(var1.get_bounds()[0], var1.get_bounds()[1], (init_samples,1))
X2 = np.random.uniform(var2.get_bounds()[0], var2.get_bounds()[1], (init_samples,1))

X_init = np.hstack((X1, X2))
Y_init = f(X_init[:,0], X_init[:,1])

# Dense grid of points within bounds
X1 = np.linspace(var1.get_bounds()[0], var1.get_bounds()[1], 1000)
X2 = np.linspace(var2.get_bounds()[0], var2.get_bounds()[1], 1000)
X_plot, Y_plot = np.meshgrid(X1, X2)
Z_plot = f(X_plot, Y_plot)

fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
ax.plot_surface(X_plot, Y_plot, Z_plot)
plt.show()       

#%%

# Initialize samples
X_sample = X_init
Y_sample = Y_init.reshape(-1,1)

kernel = GPy.kern.Matern32(input_dim=2, ARD=True)
gpr = GPy.models.GPRegression(X_sample, normalize(Y_sample), kernel)

acqu = EI(gpr, jitter=0.0)

# Build dounds for optimizer
# Tmp since the model in not integrated
bounds = np.array([var1.get_bounds(),
                   var2.get_bounds()])
    
opt = OptimizerLBRGS()
opt.set_bounds(bounds)

#%%

# Number of iterations
n_iter = 40

for i in tqdm(range(n_iter)):
    
    gpr.optimize_restarts(verbose=False)
    
    X_next = None
    y_max = np.inf

    num_anchor = 5
    X1 = np.linspace(var1.get_bounds()[0], var1.get_bounds()[1], 200)
    X2 = np.linspace(var2.get_bounds()[0], var2.get_bounds()[1], 200)
    x1, x2 = np.meshgrid(X1, X2)
    X_starts = np.hstack((x1.reshape(200*200,1),x2.reshape(200*200,1)))
    scores = acqu.get_acquisition(X_starts).flatten()
    idx = np.argsort(scores)[:min(len(scores),num_anchor)]
    anchor_points = X_starts[idx,:]
    
    for x0 in anchor_points:
        #x, y = opt.optimize(x0, f=acqu.get_acquisition, df=acqu.get_acquisition_grad)
        x, y = opt.optimize(x0, f=acqu.get_acquisition)
        if y < y_max:
            X_next = x
            y_max = y
    
    Y_next = f(X_next[:,0], X_next[:,1])
    
    # Add sample to previous samples
    X_sample = np.vstack((X_sample, X_next))
    Y_sample = np.vstack((Y_sample, Y_next))

    gpr.set_XY(X_sample, normalize(Y_sample))

plot_acquisition(gpr, bounds, acqu, X_next)
plot_approximation_2d(gpr, bounds, acqu, X_sample, X_next)
