#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 30 11:59:30 2018

@author: Stefano Toniolo
"""

import sys
sys.path.append("..") # Adds higher directory to python modules path.

import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm
from scipy.optimize import minimize
from scipy.special import erfc
import scipy.optimize

#from sklearn.gaussian_process import GaussianProcessRegressor
#from sklearn.gaussian_process.kernels import ConstantKernel, Matern

import GPy
from acquisition.ei import EI
from acquisition.lcb import LCB
from acquisition.ucb import UCB
from optimizer.optimizer import OptimizerLBRGS, OptimizerDE
from utils.plot import plot_approximation, plot_acquisition

def f(X, noise=0.0):
    return -np.sin(3*X) - X**2 + 0.7*X + noise * np.random.randn(*X.shape)

def propose_location(acquisition, X_sample, Y_sample, gpr, bounds, n_restarts=25):
    ''' Proposes the next sampling point by optimizing the acquisition function. Args: acquisition: Acquisition function. X_sample: Sample locations (n x d). Y_sample: Sample values (n x 1). gpr: A GaussianProcessRegressor fitted to samples. Returns: Location of the acquisition function maximum. '''
    dim = X_sample.shape[1]
    min_val = 1
    min_x = None
    
    def min_obj(x0):
        fmin = -acquisition.get_acquisition(x0.reshape(-1, dim))
        return fmin
    
    # Find the best optimum by starting from n_restart different random points.
    for x0 in np.linspace(bounds[:,0], bounds[:,1], n_restarts).reshape(-1,1):
        res = minimize(min_obj, x0=x0, bounds=bounds, method='L-BFGS-B')        
        
        if res.fun < min_val:
            min_val = res.fun
            min_x = res.x      
        
    return np.clip(min_x.reshape(-1, 1), bounds[:,0], bounds[:,1])     

def greedy(acquisition, bounds, n_samples=100):
    X = np.linspace(bounds[:,0], bounds[:,1], n_samples).reshape(-1,1)
    Y = acquisition.get_acquisition(X)
    idx = np.argmax(Y)
    return X[idx].reshape(-1,1)

#%%

bounds = np.array([[-1.0, 2.0]])
noise = 0.0

X_init = np.random.uniform(bounds[:,0], bounds[:,1], (2,1))
Y_init = f(X_init, noise)

# Dense grid of points within bounds
X = np.arange(bounds[:, 0], bounds[:, 1], 0.01).reshape(-1, 1)
Y = f(X)

# =============================================================================
# # Plot optimization objective with noise level
# plt.plot(X, Y, 'y--', lw=2, label='Noise-free objective')
# plt.plot(X, f(X), 'bx', lw=1, alpha=0.1, label='Noisy samples')
# plt.plot(X_init, Y_init, 'kx', mew=3, label='Initial samples')
# plt.legend();
# =============================================================================

# Initialize samples
X_sample = X_init
Y_sample = Y_init

# Gaussian process with Mat??rn kernel as surrogate model
#m52 = ConstantKernel(1.0) * Matern(length_scale=1.0, nu=2.5)
#gpr = GaussianProcessRegressor(kernel=m52, alpha=noise**2)

kernel = GPy.kern.Matern32(input_dim=1, ARD=True)
gpr = GPy.models.GPRegression(X_sample, Y_sample, kernel)

acqu = EI(gpr, jitter=0.0)
#acqu = LCB(gpr)
#acqu = UCB(gpr)

opt = OptimizerLBRGS()
opt.set_bounds(bounds)
#opt = OptimizerDE(bounds)

# Number of iterations
n_iter = 10

plt.figure(figsize=(12, n_iter * 3))
plt.subplots_adjust(hspace=0.4)
acq = np.empty((0,2))

for i in range(n_iter):
    gpr.optimize_restarts()

# =============================================================================
#     # Obtain next sampling point from the acquisition function (expected_improvement)
#     X_next = propose_location(acqu, X_sample, Y_sample, gpr, bounds)
#     X_next = greedy(acqu, bounds)
# =============================================================================
    
    X_next = None
    y_max = np.inf
    
# =============================================================================
#     X_start = gpr.X
#     Y_start = gpr.Y
#     lower = np.array([0])
#     upper = np.array([1])
#     n_restarts = 10
#     
#     starts = np.random.uniform(lower, upper, size=(int(n_restarts * 0.5),1))
#     rand_incs = np.array([np.random.normal(loc=X_start[np.argmax(Y_start)], 
#                                            scale=np.ones([lower.shape[0]]) * 0.5) 
#         for _ in range(int(n_restarts * 0.5))])
#         
#     starts = np.append(starts, rand_incs, axis=0)
# =============================================================================
    
    num_anchor = 5
    X_starts = np.random.uniform(bounds[:,0], bounds[:,1], (1000,1))
    scores = acqu.get_acquisition(X_starts).flatten()
    anchor_points = X_starts[np.argsort(scores)[:min(len(scores),num_anchor)], :]
    
    #for x0 in np.random.uniform(bounds[:,0], bounds[:,1], 25):
    for x0 in anchor_points:
        x, y = opt.optimize(x0, f=acqu.get_acquisition, df=acqu.get_acquisition_grad)
        #x, y = opt.optimize(x0, f=acqu.get_acquisition)
        if y < y_max:
            X_next = x
            y_max = y
    
    # Obtain next noisy sample from the objective function
    Y_next = f(X_next, noise)
    
    # Plot samples, surrogate function, noise-free objective and next sampling location
    plt.subplot(n_iter, 2, 2 * i + 1)
    plot_approximation(gpr, X, Y, X_sample, Y_sample, X_next, show_legend=i==0)
    plt.title(f'Iteration {i+1}')

    plt.subplot(n_iter, 2, 2 * i + 2)
    plot_acquisition(gpr, bounds, acqu, X_next, show_legend=i==0)
    
    #test = acqu.get_acquisition(X)
    #idx = np.argmax(test)
    #acq = np.append(acq, [[X[idx,0], test[idx,0]]], axis=0)
    
    # Add sample to previous samples
    X_sample = np.vstack((X_sample, X_next))
    Y_sample = np.vstack((Y_sample, Y_next))

    gpr.set_XY(X_sample, Y_sample)
