#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 30 11:59:30 2018

@author: Stefano Toniolo
"""

import sys
sys.path.append("..") # Adds higher directory to python modules path.

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from tqdm import tqdm

import GPy
from acquisition.ei import EI
from optimizer.LBFGSB import LBFGSB
from utils.plot import plot_approximation_2d, plot_acquisition
from utils.math import normalize
from variable.continuous_variable import ContinuousVariable
from variable.discrete_variable import DiscreteVariable
from variable.categorical_variable import CategoricalVariable
from points_generator.bruteforce_points_generator import BruteforcePointsGenerator

import seaborn as sns
sns.set_style("white")

#%%

def f(X):
    n = X.shape[0]
    fval = (X * np.sin(X) + 0.1 * X).sum(axis=1) 
    return -fval.reshape(n,1)

def generate_samples(fvars, anchor_points):
    X_init = np.empty((anchor_points, 0))
    for fvar in fvars:
        if fvar.get_type() == 'continuous':
            x = np.random.uniform(fvar.get_domain()[0], fvar.get_domain()[1], (anchor_points,1))
        elif fvar.get_type() == 'discrete':
            x = np.random.choice(fvar.get_domain(), (anchor_points,1))
        elif fvar.get_type() == 'categorical':
            x = fvar.get_one_hot(np.random.choice(fvar.get_domain(), (anchor_points,1)))
        else:
            raise ValueError('YOLO')
        X_init = np.column_stack((X_init, x))  
    return X_init

def generate_samples_v2(fvars, anchor_points):
    # search for categorical
    fvars_cat = []
    fvars_not_cat = []
    for fvar in fvars:
        if fvar.get_type() == 'categorical':
            fvars_cat.append(fvar)
        else:
            fvars_not_cat.append(fvar)
    
    # generate non cat samples
    X_init_not_cat = np.empty((anchor_points, 0))
    for fvar in fvars_not_cat:
        if fvar.get_type() == 'continuous':
            x = np.random.uniform(fvar.get_bounds()[0], fvar.get_bounds()[1], (anchor_points,1))
        elif fvar.get_type() == 'discrete':
            x = np.random.choice(fvar.get_domain(), (anchor_points,1))  
        else:
            raise ValueError('YOLO')
        X_init_not_cat = np.column_stack((X_init, x))
        
    # add categorical combinations
    # TODO
        
        
    return X_init

def get_rounded_opt(fvars, X):
    X_round = np.zeros_like(X)
    idx = 0
    for fvar in fvars:
        dim = fvar.get_dim()
        X_round[:,idx:idx+dim] = fvar.get_rounded(X[:,idx:idx+dim])
        idx = idx + dim
    return X_round

def normalize_X(X, fvars):
    X_norm = np.empty_like(X)
    idx = 0
    for fvar in fvars:
        if fvar.get_type() == 'categorical':
            X_norm[:,idx:idx+fvar.get_dim()] = X[:, idx:idx+fvar.get_dim()]
            idx = idx + fvar.get_dim()
        else:
            col = X[:,idx]
            bounds = fvar.get_bounds()
            X_norm[:,idx] = (col - bounds[0]) / (bounds[1] - bounds[0])
            idx = idx + 1
    return X_norm


#%%

# =============================================================================
# fvars = [ContinuousVariable('var1', np.array([-5.0,5.0])),
#          ContinuousVariable('var2', np.array([-5.0,5.0])),
#          ContinuousVariable('var3', np.array([-5.0,5.0])),
#          DiscreteVariable('var4', np.array([3,8,9])),
#          DiscreteVariable('var5', np.array([3,8,9])),
#          CategoricalVariable('var6', np.array([0,1,2])),
#          ContinuousVariable('var7', np.array([-1.0,2.0]))]
# =============================================================================

fvars = [ContinuousVariable('var3', np.array([-5.0,5.0])),
         DiscreteVariable('var5', np.array([3,8,9])),
         CategoricalVariable('var6', np.array([0,1,2])),
         CategoricalVariable('var7', np.array([0,1]))]

bounds = np.empty((0,2))
for fvar in fvars:
    bounds = np.vstack((bounds, fvar.get_bounds()))

init_samples = 5

gen = BruteforcePointsGenerator(fvars)
X, X_sample = gen.generate_points(5)

# Initialize samples
#X_sample = generate_samples(fvars, init_samples)
Y_sample = f(X)

#%%

kernel = GPy.kern.Matern32(input_dim=X_sample.shape[1], ARD=True)
gpr = GPy.models.GPRegression(normalize_X(X_sample, fvars), Y_sample, kernel)

acqu = EI(gpr, jitter=0.0)
    
opt = LBFGSB()

norm_bounds = np.zeros_like(bounds)
norm_bounds[:,1] = np.ones((bounds.shape[0]))
opt.set_bounds(norm_bounds)

#%%

# Number of iterations
n_iter = 20

for i in tqdm(range(n_iter)):
    
    gpr.optimize_restarts(verbose=False)
    
    X_next = None
    y_max = np.inf

    num_anchor = 5
    X, X_start = gen.generate_points(num_anchor)
    
# =============================================================================
#     scores = acqu.get_acquisition(X_starts).flatten()
#     idx = np.argsort(scores)[:min(len(scores),num_anchor)]
#     anchor_points = X_starts[idx,:]
# =============================================================================
    
    for i, x0 in enumerate(X_start):
        x, y = opt.optimize(x0, f=acqu.get_acquisition, df=acqu.get_acquisition_grad)
        #x, y = opt.optimize(x0, f=acqu.get_acquisition)
        if y < y_max:
            x = get_rounded_opt(fvars, x)
            X_next = i
            y_max = y
    
    Y_next = f(np.atleast_2d(X[i]))
    
    # Add sample to previous samples
    X_sample = np.vstack((X_sample, X_start[i]))
    Y_sample = np.vstack((Y_sample, Y_next))

    gpr.set_XY(normalize_X(X_sample, fvars), Y_sample)
    
print(Y_sample.min())

