#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 14 22:53:52 2019

@author: Stefano Toniolo
"""

import unittest
import numpy as np
import GPy.kern as k
import core.BO as BO

def six_humps(X):
    ''' 
    Six humps
     - Global min: Y = -1.0316
                   X = [[-3,3],[-2,2]]
    '''
    x2 = np.power(X[:,0],2)
    x4 = np.power(X[:,0],4)
    y2 = np.power(X[:,1],2)  
    fval = (4.0 - 2.1 * x2 + (x4 / 3.0)) * x2 + X[:,0]*X[:,1] + (-4.0 + 4.0 * y2) * y2
    return fval.reshape(-1,1) 

class TestSixHumps(unittest.TestCase):
    
    def setUp(self):
        self.fvar = [{'name':'var1', 'type':'continuous', 'domain': (-3,3)},
                     {'name':'var2', 'type':'continuous', 'domain': (-2,2)}]
        self.kernel = k.Matern52(input_dim=2)
        self.epsilon = 0.1
        self.run_iter = 20
        self.global_miny = -1.0316
    
    def test_bo_ei(self):
        bo = BO.BayesianOptimization(
                self.fvar, [six_humps], kernel=self.kernel, normalize_Y=True,
                acquisition_type='EI', acquisition_optimizer_type='lbfgsb',
                samples_generator='random', evaluator_type='sequential_batch')
        bo.run_optimization(self.run_iter)
        
        miny = np.min(bo.Y)
        print("Min found: {0}".format(np.min(bo.Y)))
               
        self.assertTrue(self.global_miny - self.epsilon <= miny <= self.global_miny + self.epsilon,
                        'Min Y not found')
        
    def test_bo_eimc(self):
        bo = BO.BayesianOptimization(
                self.fvar, [six_humps], kernel=self.kernel, normalize_Y=True,
                acquisition_type='EIMC', acquisition_optimizer_type='lbfgsb',
                samples_generator='random', evaluator_type='sequential_batch')
        bo.run_optimization(self.run_iter)
        
        miny = np.min(bo.Y)
        print("Min found: {0}".format(np.min(bo.Y)))
        
        self.assertTrue(self.global_miny - self.epsilon <= miny <= self.global_miny + self.epsilon,
                        'Min Y not found')

if __name__ == '__main__':
    unittest.main()