import unittest
import numpy as np
import GPy.kern as k
import core.BO as BO

def six_humps_cat(X):
    ''' 
    Six humps custom with categoricals
     - Global min: Y = -1.0316+0.25+0.4=-0.3816
                   X = [[-3,3],[-2,2],[1,2,...,N],[1,2,...,N]]
    '''
    x2 = np.power(X[:,0],2)
    x4 = np.power(X[:,0],4)
    y2 = np.power(X[:,1],2)  
    fval = (4.0 - 2.1 * x2 + (x4 / 3.0)) * x2 + X[:,0]*X[:,1] + (-4.0 + 4.0 * y2) * y2+X[:,2]*0.25+X[:,3]*0.4
    return fval.reshape(-1,1) 

class TestSixHumpsCat(unittest.TestCase):
    
    def setUp(self):
        self.fvar = [{'name':'var1', 'type':'continuous', 'domain': (-3,3)},
                     {'name':'var2', 'type':'continuous', 'domain': (-2,2)},
                     {'name':'var3', 'type':'categorical', 'domain': (1,2,3)},
                     {'name':'var4', 'type':'categorical', 'domain': (1,2,3)}]
        self.kernel = k.Matern52(input_dim=8)
        self.epsilon = 0.1
        self.run_iter = 20
        self.global_miny = -1.0316+0.25+0.4
    
    def test_bo_ei(self):
        bo = BO.BayesianOptimization(
                self.fvar, [six_humps_cat], kernel=self.kernel, normalize_Y=True,
                acquisition_type='EI', acquisition_optimizer_type='lbfgsb',
                samples_generator='bruteforce', evaluator_type='sequential_batch')
        bo.run_optimization(self.run_iter)
        
        miny = np.min(bo.Y[0])
        print("Min found: {0}".format(np.min(bo.Y[0])))
        print("x: {0}".format(bo.X[0][np.argmin(bo.Y[0]),:]))  
        
        self.assertTrue(np.abs(self.global_miny - miny)<=self.epsilon,
                        'Min Y not found')
        
    def test_bo_ei_no_bruteforce(self):
        bo = BO.BayesianOptimization(
                self.fvar, [six_humps_cat], kernel=self.kernel, normalize_Y=True,
                acquisition_type='EI', acquisition_optimizer_type='lbfgsb',
                samples_generator='random', evaluator_type='sequential_batch')
        bo.run_optimization(self.run_iter)
        
        miny = np.min(bo.Y[0])
        print("Min found: {0}".format(np.min(bo.Y[0])))
        print("x: {0}".format(bo.X[0][np.argmin(bo.Y[0]),:]))       
        
        self.assertTrue(np.abs(self.global_miny - miny)<=self.epsilon,
                        'Min Y not found')
        
    def test_bo_eimc(self):
        bo = BO.BayesianOptimization(
                self.fvar, [six_humps_cat], kernel=self.kernel, normalize_Y=True,
                acquisition_type='EIMC', acquisition_optimizer_type='lbfgsb',
                samples_generator='bruteforce', evaluator_type='sequential_batch')
        bo.run_optimization(self.run_iter)
        
        miny = np.min(bo.Y[0])
        print("Min found: {0}".format(np.min(bo.Y[0])))
        print("x: {0}".format(bo.X[0][np.argmin(bo.Y[0]),:])) 
        
        self.assertTrue(self.global_miny - self.epsilon <= miny <= self.global_miny + self.epsilon,
                        'Min Y not found')
        
    def test_bo_eimc_no_bruteforce(self):
        bo = BO.BayesianOptimization(
                self.fvar, [six_humps_cat], kernel=self.kernel, normalize_Y=True,
                acquisition_type='EIMC', acquisition_optimizer_type='lbfgsb',
                samples_generator='random', evaluator_type='sequential_batch')
        bo.run_optimization(self.run_iter)
        
        miny = np.min(bo.Y[0])
        print("Min found: {0}".format(np.min(bo.Y[0])))
        print("x: {0}".format(bo.X[0][np.argmin(bo.Y[0]),:]))    
        
        self.assertTrue(np.abs(self.global_miny - miny)<=self.epsilon,
                        'Min Y not found')

if __name__ == '__main__':
    unittest.main()